///////////////////////////////////////////////////////////////////////////////////////////////
//
// EventUtils.pas
// --------------------------
// Changed:   2003-02-06
// Maintain:  Michael Vinter   |   mv@logicnet.dk
//
// Last change:
//   SetSeparatorVissible added
//
unit EventUtils;

interface

uses Windows, ShellAPI, Forms, StdCtrls, Graphics, Classes, Controls, FileUtils, Monitor,
  MemUtils, Comctrls, SysUtils;

// Get TRect with Windows taskbar
function GetTaskBarRect: TRect;
// Turn a label into a link. The Hint property must be the link target.
procedure MakeLinkLabel(LinkLabel: TLabel);
// Frees UnwantedObject when application is idle
procedure FreeOnIdle(UnwantedObject: TObject);
procedure CloseOnIdle(UnwantedForm: TCustomForm);
procedure NotifyOnIdle(AEvent: TNotifyEvent);
// Clear label captions if Label.Caption=Label.Name
procedure ClearDelphiLabelCaptions(Control: TWinControl);
// Reset anchors in all child controls to [akLeft,akTop]
procedure ResetControlAnchors(Control: TControl);
// Change visibility of a separator (or button) in a tool bar
procedure SetSeparatorVissible(var Separator: TToolButton; AVisible: Boolean; ALeft: Integer=Low(Integer));

type
  TLinkExecute = class(TComponent)
                   procedure OnClick(Sender: TObject);
                 end;

// Set caller thread priority
procedure SetPriority(Priority: TThreadPriority);

implementation

procedure SetPriority(Priority: TThreadPriority);
const
  Priorities: array [TThreadPriority] of Integer =
   (THREAD_PRIORITY_IDLE, THREAD_PRIORITY_LOWEST, THREAD_PRIORITY_BELOW_NORMAL,
    THREAD_PRIORITY_NORMAL, THREAD_PRIORITY_ABOVE_NORMAL,
    THREAD_PRIORITY_HIGHEST, THREAD_PRIORITY_TIME_CRITICAL);
begin
  SetThreadPriority(GetCurrentThread,Priorities[Priority]);
end;

procedure SetSeparatorVissible(var Separator: TToolButton; AVisible: Boolean; ALeft: Integer);
var
  New : TToolButton;
begin
  with Separator do
  if AVisible<>Visible then
  begin
    New:=TToolButton.Create(Owner);
    New.Action:=Action;
    New.AutoSize:=False;
    New.Width:=Width;
    New.Height:=Height;
    if ALeft=Low(Integer) then New.Left:=Left
    else New.Left:=ALeft;
    New.Visible:=AVisible;
    New.Style:=Style;
    New.Parent:=Parent;
    Free;
    Separator:=New;
  end;
end;

procedure ClearDelphiLabelCaptions(Control: TWinControl);
var
  I : Integer;
begin
  with Control do
  for I:=0 to ControlCount-1 do
    if Controls[I] is TLabel then
    begin
      if Controls[I].Name=TLabel(Controls[I]).Caption then TLabel(Controls[I]).Caption:='';
    end
    else if Controls[I] is TWinControl then ClearDelphiLabelCaptions(TWinControl(Controls[I]));
end;

procedure ResetControlAnchors(Control: TControl);
var
  I : Integer;
begin
  Control.Anchors:=[akLeft,akTop];
  if Control is TWinControl then
    with TWinControl(Control) do
      for I:=0 to ControlCount-1 do ResetControlAnchors(Controls[I]);
end;

//==============================================================================================================================
// MakeLinkLabel
//==============================================================================================================================

procedure TLinkExecute.OnClick(Sender: TObject);
begin
  ExecuteFile(TControl(Sender).Hint,'','');
end;

// Turn a label into a link. The Hint property must be the link target.
procedure MakeLinkLabel(LinkLabel: TLabel);
begin
  LinkLabel.Font.Color:=clBlue;
  LinkLabel.Font.Style:=LinkLabel.Font.Style+[fsUnderline];
  LinkLabel.Cursor:=crHandPoint;
  LinkLabel.OnClick:=TLinkExecute.Create(LinkLabel).OnClick;
end;

//==============================================================================================================================
// TActionOnIdle
//==============================================================================================================================

type
  TActionOnIdle = class(TMonitorObject)
                    Original : TIdleEvent;
                    Obj : TObject;
                    Event : TNotifyEvent;
                    procedure DoFree(Sender: TObject; var Done: Boolean);
                    procedure DoClose(Sender: TObject; var Done: Boolean);
                    procedure Notify(Sender: TObject; var Done: Boolean);
                  end;

procedure TActionOnIdle.DoFree(Sender: TObject; var Done: Boolean);
begin
  Application.OnIdle:=Original;
  try
    Obj.Free;
  finally
    if Assigned(Original) then Original(Sender,Done);
    Destroy;
  end;
end;

procedure TActionOnIdle.DoClose(Sender: TObject; var Done: Boolean);
begin
  Application.OnIdle:=Original;
  try
    TCustomForm(Obj).Close;
  finally
    if Assigned(Original) then Original(Sender,Done);
    Destroy;
  end;
end;

procedure TActionOnIdle.Notify(Sender: TObject; var Done: Boolean);
begin
  Application.OnIdle:=Original;
  try
    Event(nil);
  finally
    if Assigned(Original) then Original(Sender,Done);
    Destroy;
  end;
end;

procedure FreeOnIdle(UnwantedObject: TObject);
begin
  with TActionOnIdle.Create do
  begin
    Original:=Application.OnIdle;
    Application.OnIdle:=DoFree;
    Obj:=UnwantedObject;
  end;
end;

procedure CloseOnIdle(UnwantedForm: TCustomForm);
begin
  with TActionOnIdle.Create do
  begin
    Original:=Application.OnIdle;
    Application.OnIdle:=DoClose;
    Obj:=UnwantedForm;
  end;
end;

procedure NotifyOnIdle(AEvent: TNotifyEvent);
begin
  with TActionOnIdle.Create do
  begin
    Original:=Application.OnIdle;
    Application.OnIdle:=Notify;
    Event:=AEvent;
  end;
end;

function GetTaskBarRect: TRect;
var
  BarData: TAppBarData;
begin
  ZeroMem(Result,SizeOf(Result));
  BarData.cbSize:=SizeOf(barData);
  BarData.hWnd:=FindWindow('Shell_TrayWnd',nil);
  SHAppBarMessage(ABM_GETTASKBARPOS,BarData);
  Result:=BarData.rc;
end;

end.

